import pytest

import awkward as ak

from fast_carpenter.weights import get_unweighted_increment, get_weighted_increment, extract_weights


@pytest.fixture
def event_weights():
    weights = ak.Array(
        [
            [1, 2, 3],  # event level weights
            [2, 5, 7],  # event level weights
            # [[0.4], [0.5, 0.6], [0.7, 0.8, 0.9]],  # object, e.g. muon, level weights
        ]
    )
    # TODO: to support awkward arrays as weights, simply remove the "to_numpy"
    return weights.to_numpy()


@pytest.fixture
def event_mask():
    return ak.Array([True, False, True])


def test_unweighted_increment(event_weights):
    inc = get_unweighted_increment(event_weights, None)
    assert inc == len(event_weights[0])


def test_weighted_increment(event_weights):
    inc = get_weighted_increment(event_weights, None)
    assert ak.all(inc == [6, 14])


def test_extract_weights(fake_data_events):
    weights = extract_weights(fake_data_events, ["EventWeight"])
    assert len(weights) == 1
    assert len(weights[0]) == len(fake_data_events)


def test_extract_weights_no_weights(fake_data_events):
    weights = extract_weights(fake_data_events, [])
    assert len(weights) == 0



"""
Provides a common interface to select a backend

Each backend is wrapped in a function so that it is only imported if requested
"""


def get_alphatwirl():
    from . import _alphatwirl
    return _alphatwirl


def get_coffea():
    from . import coffea
    return coffea


KNOWN_BACKENDS = {
    "multiprocessing": get_alphatwirl,
    "htcondor": get_alphatwirl,
    "sge": get_alphatwirl,
    "coffea:local": get_coffea,
    "coffea:parsl": get_coffea,
    "coffea:dask": get_coffea,
}

KNOW_BACKENDS_NAMES = ", ".join(list(KNOWN_BACKENDS.keys()))


def get_backend(name):
    if name not in KNOWN_BACKENDS:
        raise ValueError(f"Unknown backend requested, '{name}'. Known backends: {KNOW_BACKENDS_NAMES}")
    return KNOWN_BACKENDS[name]()



"""
Summarize the data by producing binned and possibly weighted counts of the data.
"""
import os
import re
import numpy as np
import pandas as pd
from pandas.api.types import is_object_dtype
from . import binning_config as cfg

from fast_carpenter.tree_adapter import ArrayMethods


class Collector():
    valid_ext = {'xlsx': 'excel', 'h5': 'hdf', 'msg': 'msgpack', 'dta': 'stata', 'pkl': 'pickle', 'p': 'pickle'}

    def __init__(self, filename, dataset_col, binnings, file_format):
        self.filename = filename
        self.dataset_col = dataset_col
        self.binnings = binnings
        self.file_format = file_format

    def collect(self, dataset_readers_list, doReturn=True, writeFiles=True):
        if len(dataset_readers_list) == 0:
            if doReturn:
                return pd.DataFrame()
            else:
                return None

        output = self._prepare_output(dataset_readers_list)

        if writeFiles:
            for file_dict in self.file_format:
                file_ext = file_dict.pop('extension', None)
                save_func = file_ext.split('.')[1]
                if save_func in Collector.valid_ext:
                    save_func = Collector.valid_ext[save_func]
                try:
                    getattr(output, "to_%s" % save_func)(self.filename + file_ext, **file_dict)
                except AttributeError as err:
                    print("Incorrect file format: %s" % err)
                except TypeError as err:
                    print("Incorrect args: %s" % err)

        if doReturn:
            return output

    def _prepare_output(self, dataset_readers_list):
        return combined_dataframes(dataset_readers_list,
                                   self.dataset_col,
                                   binnings=self.binnings)


def combined_dataframes(dataset_readers_list, dataset_col, binnings=None):
    dataset_readers_list = [(d, [r.contents for r in readers])
                            for d, readers in dataset_readers_list if readers]
    if not dataset_readers_list:
        return None

    if dataset_col:
        output = _merge_dataframes(dataset_readers_list)
    else:
        output = _sum_dataframes(dataset_readers_list)
    if binnings:
        output = densify_dataframe(output, binnings)
    return output


def _merge_dataframes(dataset_readers_list):
    all_dfs = []
    keys = []
    for dataset, readers in dataset_readers_list:
        dataset_df = readers[0]
        for df in readers[1:]:
            if df is None or df.empty:
                continue
            dataset_df = dataset_df.add(df, fill_value=0.)
        if dataset_df is None or dataset_df.empty:
            continue
        all_dfs.append(dataset_df)
        keys.append(dataset)
    if all_dfs:
        final_df = pd.concat(all_dfs, keys=keys, names=['dataset'], sort=True)
    else:
        final_df = pd.DataFrame()

    return final_df


def _sum_dataframes(dataset_readers_list):
    final_df = None
    for _, readers in dataset_readers_list:
        for df in readers:
            if df is None:
                continue
            if final_df is None:
                final_df = df
                continue
            final_df = final_df.add(df, fill_value=0.)
    return final_df


def densify_dataframe(in_df, binnings):
    in_index = in_df.index
    index_values = []
    for dim in in_index.names:
        bins = binnings.get(dim, None)
        if bins is None:
            index_values.append(in_index.unique(dim))
            continue
        index_values.append(bins)
    out_index = pd.MultiIndex.from_product(index_values, names=in_index.names)
    out_df = in_df.reindex(index=out_index, copy=False)
    return out_df


class BinnedDataframe():
    """Produces a binned dataframe (a multi-dimensional histogram).

    def __init__(self, name, out_dir, binning, weights=None, dataset_col=False):

    Parameters:
      binning (list[dict]): A list of dictionaries describing the variables to
        bin on, and how they should be binned.  Each of these dictionaries can
        contain the following:

        +-----------+----------------+--------------------------------------------------------------+
        | Parameter | Default        | Description                                                  |
        +===========+================+==============================================================+
        | ``in``    |                | The name of the attribute on the event to use.               |
        +-----------+----------------+--------------------------------------------------------------+
        | ``out``   | same as ``in`` | The name of the column to be filled in the output dataframe. |
        +-----------+----------------+--------------------------------------------------------------+
        | ``bins``  | ``None``       | | Must be either ``None`` or a dictionary.  If a dictionary, |
        |           |                |   it must contain one of the follow sets of                  |
        |           |                | | key-value pairs:                                           |
        |           |                | |   1. ``nbins``, ``low``, ``high``: which are used to       |
        |           |                |     produce a list of bin edges equivalent to:               |
        |           |                | |      ``numpy.linspace(low, high, nbins + 1)``              |
        |           |                | |   2. ``edges``: which is treated as the list of bin        |
        |           |                |     edges directly.                                          |
        |           |                | | If set to ``None``, then the input variable is assumed     |
        |           |                |   to already be categorical (ie. binned or discrete)         |
        +-----------+----------------+--------------------------------------------------------------+

      weights (str or list[str], dict[str, str]): How to weight events in the
        output table.  Must be either a single variable, a list of
        variables, or a dictionary where the values are variables in the data and
        keys are the column names that these weights should be called in the
        output tables.
      file_format (str or list[str], dict[str, str]): determines the file format to
        use to save the binned dataframe to disk.  Should be either a) a string with
        the file format, b) a dict containing the keyword `extension` to give the file
        format and then all other keyword-argument pairs are passed on to the
        corresponding pandas function, or c) a list of values matching a) or b).
      dataset_col (bool): adds an extra binning column with the name for each dataset.
      pad_missing (bool): If ``False``, any bins that don't contain data are
        excluded from the stored dataframe.  Leaving this ``False`` can save
        some disk-space and improve processing time, particularly if the bins are
        only very sparsely filled.
      observed (bool): If ``False`` bins in the dataframe will only be filled
        if their are datapoints contained within them.  Otherwise, depending on
        the binning specification for each dimension, all bins for that
        dimension will be present.  Use `pad_missing: true` to force all bins
        to be present.

    Other Parameters:
      name (str):  The name of this stage (handled automatically by fast-flow)
      out_dir (str):  Where to put the summary table (handled automatically by
          fast-flow)

    Raises:
      BadBinnedDataframeConfig: If there is an issue with the binning description.

    """

    def __init__(self, name, out_dir, binning, weights=None, dataset_col=True,
                 pad_missing=False, file_format=None, observed=False, weight_data=False):
        self.name = name
        self.out_dir = out_dir
        ins, outs, binnings = cfg.create_binning_list(self.name, binning)
        self._bin_dims = ins
        self.potential_inputs = set(sum((re.findall(r"\w+", dim) for dim in self._bin_dims), []))
        self._out_bin_dims = outs
        self._binnings = binnings
        self._dataset_col = dataset_col
        self._weights = cfg.create_weights(self.name, weights)
        self._pad_missing = pad_missing
        self._file_format = cfg.create_file_format(self.name, file_format)
        self._observed = observed
        self.contents = None
        self.weight_data = weight_data

    def collector(self):
        outfilename = "tbl_"
        if self._dataset_col:
            outfilename += "dataset."
        outfilename += ".".join(self._out_bin_dims)
        outfilename += "--" + self.name
        outfilename = os.path.join(self.out_dir, outfilename)
        binnings = None
        if self._pad_missing:
            binnings = dict(zip(self._out_bin_dims, self._binnings))
        return Collector(outfilename, self._dataset_col, binnings=binnings, file_format=self._file_format)

    def event(self, chunk):
        all_inputs = [key for key in chunk.tree.keys() if key in self.potential_inputs]
        if chunk.config.dataset.eventtype == "mc" or self.weight_data:
            weights = list(self._weights.values())
            all_inputs += weights
        else:
            weights = None

        data = ArrayMethods.to_pandas(chunk.tree, all_inputs)
        data = explode(data)
        if data is None or data.empty:
            return True

        binned_values = _bin_values(data, dimensions=self._bin_dims,
                                    binnings=self._binnings,
                                    weights=weights,
                                    out_weights=self._weights.keys(),
                                    out_dimensions=self._out_bin_dims,
                                    observed=self._observed)
        if self.contents is None:
            self.contents = binned_values
        else:
            self.contents = self.contents.add(binned_values, fill_value=0)

        return True

    def merge(self, rhs):
        if rhs.contents is None or len(rhs.contents) == 0:
            return
        if self.contents is None:
            self.contents = rhs.contents
            return
        self.contents = self.contents.add(rhs.contents, fill_value=0)


count_label = "n"
weight_labels = ("sumw", "sumw2")


def _make_column_labels(weights):
    labels = [weight + ":" + label for label in weight_labels for weight in weights]
    return [count_label] + labels


def _bin_values(data, dimensions, binnings, weights, out_dimensions=None, out_weights=None, observed=True):
    if not out_dimensions:
        out_dimensions = dimensions
    if not out_weights:
        out_weights = weights

    final_bin_dims = []
    for dimension, binning in zip(dimensions, binnings):
        if binning is None:
            final_bin_dims.append(dimension)
            continue
        out_dimension = dimension + "_bins"
        data[out_dimension] = pd.cut(data.eval(dimension, engine='numexpr'), binning, right=False)
        final_bin_dims.append(out_dimension)

    if weights:
        weight_sq_dims = [w + "_squared" for w in weights]
        data[weight_sq_dims] = data[weights] ** 2

    bins = data.groupby(final_bin_dims, observed=observed)
    counts = bins[data.columns[0]].count()

    if weights:
        sums = bins[weights].sum()
        sum_sqs = bins[weight_sq_dims].sum()
        histogram = pd.concat([counts, sums, sum_sqs], axis="columns")
        histogram.columns = _make_column_labels(out_weights)
    else:
        histogram = counts.to_frame(count_label)

    histogram.index.set_names(out_dimensions, inplace=True)
    return histogram


_explodable_types = (tuple, list, np.ndarray)


def explode(df):
    """
    Based on this answer:
    https://stackoverflow.com/questions/12680754/split-explode-pandas\
    -dataframe-string-entry-to-separate-rows/40449726#40449726
    """
    if df is None or df.empty:
        return df

    # get the list columns
    lst_cols = [col for col, dtype in df.dtypes.items() if is_object_dtype(dtype)]
    # Be more specific about which objects are ok
    lst_cols = [col for col in lst_cols if isinstance(df[col].iloc[0], _explodable_types)]
    if not lst_cols:
        return df

    # all columns except `lst_cols`
    idx_cols = df.columns.difference(lst_cols)

    # check all lists have same length
    lens = pd.DataFrame({col: df[col].str.len() for col in lst_cols})
    different_length = (lens.nunique(axis=1) > 1).any()
    if different_length:
        raise ValueError("Cannot bin multiple arrays with different jaggedness")
    lens = lens[lst_cols[0]]

    # create "exploded" DF
    flattened = {col: df.loc[lens > 0, col].values for col in lst_cols}
    flattened = {col: sum(map(list, vals), []) for col, vals in flattened.items()}
    res = pd.DataFrame({col: np.repeat(df[col].values, lens) for col in idx_cols})
    res = res.assign(**flattened)

    # Check that rows are fully "exploded"
    return explode(res)



from abc import ABC, abstractmethod
from typing import Any, Dict, List


class DataImportBase(ABC):
    """
    This Abstract Base Class is the base class for all data import classes.
    """
    config: Dict[str, Any]

    def __init__(self, config: Dict[str, Any]) -> None:
        self.config = config

    @abstractmethod
    def _process_config(self) -> None:
        """
        Process the configuration.
        """
        pass

    @abstractmethod
    def open(self, paths: List[str]) -> Any:
        """
        This method is called by the importer to open the files.
        """
        pass



from collections import namedtuple
import numpy as np

FakeEventRange = namedtuple("FakeEventRange", "start_entry stop_entry entries_in_block")


class Namespace():
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class FakeBEEvent(object):
    def __init__(self, tree, eventtype):
        self.tree = tree
        self.config = Namespace(dataset=Namespace(eventtype=eventtype))

    def __len__(self):
        return len(self.tree)

    def count_nonzero(self):
        return self.tree.count_nonzero()

    def arrays(self, array_names, library="np", outputtype=list):
        return [self.tree[name] for name in array_names]


class FakeTree(dict):
    length: int = 101

    def __init__(self, length=101):
        super(FakeTree, self).__init__(
            NMuon=np.linspace(0, 5., length),
            NElectron=np.linspace(0, 10, length),
            NJet=np.linspace(2, -18, length),
        )
        self.length = length

    def __len__(self):
        return self.length

    def arrays(self, array_names, library="np", outputtype=list):
        return [self[name] for name in array_names]



from typing import Any, Dict, List

from ._base import DataImportBase


class Uproot3DataImport(DataImportBase):
    """
    This class is a wrapper around the uproot3 library.
    """

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__(config)

    def _process_config(self):
        pass

    def open(self, paths: List[str]) -> Any:
        """
        This method is called by the importer to open the file.
        """
        import uproot3
        if len(paths) != 1:
            # TODO - support multiple paths
            raise AttributeError("Multiple paths not yet supported")
        # Try to open the tree - some machines have configured limitations
        # which prevent memmaps from begin created. Use a fallback - the
        # localsource option
        input_file = paths[0]
        try:
            rootfile = uproot3.open(input_file)
        except MemoryError:
            rootfile = uproot3.open(input_file, localsource=uproot3.FileSource.defaults)
        return rootfile



import numpy as np
import pytest
from fast_carpenter.selection.filters import Counter


@pytest.fixture
def weight_names():
    return [
        "EventWeight",
        # "MuonWeight", "ElectronWeight", "JetWeight",
    ]


@pytest.fixture
def counter(weight_names):
    return Counter(weight_names)


def test_init(weight_names, full_wrapped_tree):
    c = Counter(weight_names)
    assert c._weight_names == weight_names
    assert c.counts == (0, 0.0)
    assert c._w_counts == (0.0)


def test_increment_mc(counter, full_wrapped_tree):
    counter.increment(full_wrapped_tree, is_mc=True)
    n_events = len(full_wrapped_tree)
    expected_weighted_sum = 229.94895935058594
    # expected value is taken from numpy sum, but awkward sum is used
    # the difference is small and due to optimization
    # see https://github.com/scikit-hep/awkward-1.0/issues/1241
    assert counter._w_counts == pytest.approx(np.array([expected_weighted_sum]), 1e-4)
    assert counter.counts == (n_events, pytest.approx(expected_weighted_sum, 1e-4))


def test_increment_data(counter, full_wrapped_tree):
    counter.increment(full_wrapped_tree, is_mc=False)
    n_events = len(full_wrapped_tree)
    assert counter._w_counts == (n_events)
    assert counter.counts == (n_events, n_events)


def test_add(counter, full_wrapped_tree):
    counter.increment(full_wrapped_tree, is_mc=True)
    counter.add(counter)

    n_events = len(full_wrapped_tree)
    expected_weighted_sum = 229.94895935058594
    # expected value is taken from numpy sum, but awkward sum is used
    # the difference is small and due to optimization
    # see https://github.com/scikit-hep/awkward-1.0/issues/1241
    assert counter._w_counts == pytest.approx((expected_weighted_sum * 2,), 2e-4)
    assert counter.counts == (n_events * 2, pytest.approx(expected_weighted_sum * 2, 2e-4))


def test_increment_without_weights(full_wrapped_tree):
    counter = Counter([])
    counter.increment(full_wrapped_tree, is_mc=True)
    n_events = len(full_wrapped_tree)

    with pytest.raises(IndexError):
        assert counter._w_counts[0] == n_events
    assert counter.counts == (n_events, )



import awkward as ak


def check_data(data, n_data, n_nonzero, n_mask):
    assert len(data) == n_data
    assert data.count_nonzero() == n_nonzero
    if data.tree._mask is not None:
        assert len(data.tree._mask) == n_mask


def test_single_mask(fake_data_events, full_wrapped_tree, at_least_two_muons, tmpdir):
    full_size = len(fake_data_events)
    check_data(fake_data_events, n_data=full_size, n_nonzero=full_size, n_mask=full_size)
    at_least_two_muons.event(fake_data_events)
    check_data(fake_data_events, n_data=full_size, n_nonzero=289, n_mask=full_size)

    mask = fake_data_events.tree._mask
    assert ak.count_nonzero(mask) == 289
    assert ak.all(fake_data_events.tree["Muon_Pz"] == ak.mask(full_wrapped_tree["Muon_Pz"], mask))


def test_single_mask_twice(fake_data_events, full_wrapped_tree, at_least_two_muons, tmpdir):
    full_size = len(fake_data_events)
    check_data(fake_data_events, n_data=full_size, n_nonzero=full_size, n_mask=full_size)
    # applying the same cut twice should not change the result
    at_least_two_muons.event(fake_data_events)
    at_least_two_muons.event(fake_data_events)
    check_data(fake_data_events, n_data=full_size, n_nonzero=289, n_mask=full_size)

    mask = fake_data_events.tree._mask
    assert ak.count_nonzero(mask) == 289
    assert ak.all(fake_data_events.tree["Muon_Pz"] == ak.mask(full_wrapped_tree["Muon_Pz"], mask))


def test_complex_mask(fake_data_events, full_wrapped_tree, at_least_two_muons_plus, tmpdir):
    full_size = len(fake_data_events)
    check_data(fake_data_events, n_data=full_size, n_nonzero=full_size, n_mask=full_size)

    at_least_two_muons_plus.event(fake_data_events)

    check_data(fake_data_events, n_data=full_size, n_nonzero=2, n_mask=full_size)

    mask = fake_data_events.tree._mask
    assert ak.count_nonzero(mask) == 2
    assert ak.all(fake_data_events.tree["Muon_Pz"] == ak.mask(full_wrapped_tree["Muon_Pz"], mask))


def test_multi_mask(fake_data_events, full_wrapped_tree, at_least_two_muons, at_least_two_muons_plus, tmpdir):
    full_size = len(fake_data_events)
    check_data(fake_data_events, n_data=full_size, n_nonzero=full_size, n_mask=full_size)

    at_least_two_muons.event(fake_data_events)
    at_least_two_muons_plus.event(fake_data_events)

    check_data(fake_data_events, n_data=full_size, n_nonzero=2, n_mask=full_size)

    mask = fake_data_events.tree._mask
    assert ak.count_nonzero(mask) == 2
    assert ak.all(fake_data_events.tree["Muon_Pz"] == ak.mask(full_wrapped_tree["Muon_Pz"], mask))



from pathlib import Path
from ._base import DataImportBase
from ._uproot4 import Uproot4DataImport
from ._uproot3 import Uproot3DataImport

_DATA_IMPORT_PLUGINS = {
    "uproot4": Uproot4DataImport,
    "uproot3": Uproot3DataImport,
}


def register_data_import_plugin(plugin_name: str, plugin_class: DataImportBase) -> None:
    """
    Register a new plugin_name for data import.
    """
    _DATA_IMPORT_PLUGINS[plugin_name] = plugin_class


def _process_plugin_config(plugin_name: str, plugin_config: Path) -> None:
    """
        Process the plugin configuration file.
        Reads the "register" and "plugin_name" sections to register and configure the plugin.
    """
    if plugin_config is None:
        return
    if not plugin_config.exists():
        raise ValueError(f"Plugin config file {plugin_config} does not exist")
    if not plugin_config.is_file():
        raise ValueError(f"Plugin config file {plugin_config} is not a file")

    raise NotImplementedError("Plugin config file processing not implemented yet")


def get_data_import_plugin(plugin_name: str, plugin_config: Path) -> DataImportBase:
    """
    Get a data import plugin by name.
    """
    config = _process_plugin_config(plugin_name, plugin_config)
    if plugin_name not in _DATA_IMPORT_PLUGINS:
        raise ValueError(f"Unknown data import plugin: {plugin_name}")
    return _DATA_IMPORT_PLUGINS[plugin_name](config)



from typing import Any, List

import awkward as ak
import numpy as np

from .tree_adapter import ArrayMethods


def extract_weights(data: Any, weight_names: List[str]) -> np.ndarray:
    return ArrayMethods.arrays_as_np_array(data, weight_names)


def get_unweighted_increment(weights: Any, mask: Any) -> int:
    """
        Returns the total number of unweighted events.
        If no mask is present, this should be the smallest dimenion of weights.
        If a mask is present, this should be the number of events passing the mask (i.e. non-zero entries).
    """
    if mask is None:
        return min([len(w) for w in weights])
    elif ArrayMethods.is_bool(mask):
        return np.count_nonzero(mask)
    else:
        return len(mask)


def get_weighted_increment(weights: np.ndarray, mask: Any) -> np.ndarray:
    """
        Returns the total number of weighted events per weight category.
        If no mask is present, this will sum up all individual weights.
        If a mask is present, the sum is only performed over the entries that are not masked.
    """
    if mask is not None:
        weight_mask = [mask] * len(weights)  # repeat mask for each weight
        weights = ak.mask(weights, weight_mask)
    return ArrayMethods.sum(weights, axis=1)



# input is an uproot tree (uproot3 --> adapter V0, uproot4 --> adapter V1)
# output is a dict with the same structure as the tree
from collections import abc
from itertools import chain
import logging
from typing import Any, Callable, Dict, List, Optional, Protocol

import awkward as ak
import numpy as np

adapters: Dict[str, Callable] = {}
DEFAULT_TREE_TO_DICT_ADAPTOR = "uproot4"
logger = logging.getLogger(__name__)

LIBRARIES = {
    "awkward": ["ak", "ak.Array", "awkward"],
    "numpy": ["np", "np.ndarray", "numpy"],
    "pandas": ["pd", "pd.DataFrame", "pandas"],
}

SUPPORTED_OUTPUT_TYPES = [dict, tuple, list]


def register(name: str, adaptor_creation_func: Callable) -> None:
    """
    Register an adaptor.
    """
    adapters[name] = adaptor_creation_func


def unregister(name: str) -> None:
    """
    Unregister an adaptor.
    """
    adapters[name].pop()


class ArrayLike(Protocol):
    pass


class TreeLike(Protocol):
    pass


class TreeToDictAdaptor(abc.MutableMapping):
    """
    Provides a dict-like interface to a tree-like data object (e.g. ROOT TTree, uproot.tree, etc).
    """

    tree: Any
    aliases: Dict[str, Any]
    extra_variables: Dict[str, Any]

    def __init__(self, tree: Any, aliases: Dict[str, Any] = None) -> None:
        self.tree = tree
        self.aliases = aliases if aliases else {}
        self.extra_variables = {}

    def __getitem__(self, key: str) -> Any:
        """
        Get an item from the tree.
        Resolves aliases if defined.
        """
        return self.__m_getitem__(self.__resolve_key__(key))

    def __setitem__(self, key, value) -> None:
        """
        Creates a new branch in the tree.
        Resolves aliases if defined.
        """
        self.__m_setitem__(self.__resolve_key__(key), value)

    def __iter__(self) -> Any:
        return iter(self.tree)

    def __len__(self) -> int:
        """ Returns the number of branches in the tree. """
        return len(self.tree)

    def __delitem__(self, key) -> None:
        """ Deletes a branch from the tree. """
        self.__m_delitem__(self.__resolve_key__(key))

    def __contains__(self, key):
        return self.__m_contains__(self.__resolve_key__(key))

    def keys(self) -> List[str]:
        all_keys = chain(self.tree.keys(), self.aliases.keys(), self.extra_variables.keys())
        for key in all_keys:
            yield key

    def new_variable(self, key, value, context: Optional[Any] = None) -> None:
        if context is None:
            context = self
        if len(value) != context.num_entries:
            msg = f"New variable {key} does not have the right length: {len(value)} not {context.num_entries}"
            raise ValueError(msg)

        if key not in self:
            self.__new_variable__(self.__resolve_key__(key), value)
        else:
            raise ValueError(f"Trying to overwrite existing variable: {key}")

    def evaluate(self, expression: str) -> Any:
        """
        Evaluate a string expression using the tree as the namespace.
        """
        raise NotImplementedError()

    @property
    def num_entries(self) -> int:
        """ Returns the number of entries in the tree. """
        raise NotImplementedError()


class IndexingMixin(object):
    """
    Provides indexing support for the dict-like interface.
    """
    aliases: Dict[str, str]
    special_token_mapping: Dict[str, str] = {
        ".": "__DOT__",
    }

    def __resolve_key__(self, key):
        if key in self.aliases:
            return self.aliases[key]
        return key

    def __resolve_special_tokens__(self, key):
        new_key = key
        for token in self.special_token_mapping:
            if token in key:
                new_key = new_key.replace(token, self.special_token_mapping[token])
        return new_key


class Uproot3Methods(object):
    """
    Provides uproot3-specific methods for the dict-like interface.
    """

    def __m_getitem__(self, key):
        return self.tree.array(key)

    def __m_setitem__(self, key, value):
        self.tree.set_branch(key, value)

    def __m_delitem__(self, key):
        self.tree.drop_branch(key)

    def __m_contains__(self, key):
        return key in self.tree

    @property
    def num_entries(self) -> int:
        return self.tree.numentries

    @staticmethod
    def counts(array, **kwargs):
        if hasattr(array, "compact"):
            return array.compact().counts
        else:
            return array.counts

    def array(self, key):
        return self[key]

    @staticmethod
    def pad(array, length, **kwargs):
        return array.pad(length, **kwargs)

    @staticmethod
    def flatten(array, **kwargs):
        return array.flatten(**kwargs)

    @staticmethod
    def sum(array, **kwargs):
        return array.sum(**kwargs)

    @staticmethod
    def prod(array, **kwargs):
        return array.prod(**kwargs)

    @staticmethod
    def any(array, **kwargs):
        return array.any(**kwargs)

    @staticmethod
    def all(array, **kwargs):
        return array.all(**kwargs)

    @staticmethod
    def count_nonzero(array, **kwargs):
        return array.count_nonzero(**kwargs)

    @staticmethod
    def max(array, **kwargs):
        return array.max(**kwargs)

    @staticmethod
    def min(array, **kwargs):
        return array.min(**kwargs)

    @staticmethod
    def argmax(array, **kwargs):
        return array.argmax(**kwargs)

    @staticmethod
    def argmin(array, **kwargs):
        return array.argmin(**kwargs)

    @staticmethod
    def count_zero(array, **kwargs):
        return np.count_zero(array, **kwargs)

    @staticmethod
    def dtype(array, **kwargs):
        return array.dtype.kind

    @staticmethod
    def is_bool(array, **kwargs):
        return Uproot3Methods.dtype(array) == "b"

    @staticmethod
    def arrays_as_np_lists(data, array_names, **kwargs):
        return data.arrays(array_names, outputtype=lambda *args: np.array(args))

    @staticmethod
    def to_pandas(data, keys, flatten):
        return data.pandas.df(keys, flatten=flatten)


class Uproot4Methods(object):
    """
    Provides uproot4-specific methods for the dict-like interface.
    """

    def __m_getitem__(self, key):
        if key in self.extra_variables:
            return self.extra_variables[key]
        if hasattr(self.tree[key], "array"):
            return self.tree[key].array()
        return self.tree[key]

    def __m_setitem__(self, key, value):
        self.tree.set_branch(key, value)

    def __m_delitem__(self, key):
        self.tree.drop_branch(key)

    def __m_contains__(self, key):
        return key in self.tree.keys() or key in self.extra_variables.keys()

    def __new_variable__(self, key, value) -> None:
        key = self.__resolve_special_tokens__(key)
        self.extra_variables[key] = value

    @property
    def num_entries(self) -> int:
        try:
            return self.tree.num_entries
        except AttributeError as e:
            logger.error(f"Object of type {type(self.tree)} does not have a num_entries attribute.")
            raise e

    @staticmethod
    def arraydict_to_pandas(arraydict: Dict[str, Any]):
        """
        Converts a dictionary of arrays to a pandas DataFrame.
        """
        return ak.to_pandas(arraydict)

    def array_dict(self, keys: List[str]) -> Dict[str, Any]:
        """
        Returns a dictionary of arrays for the given keys.
        """
        extra_arrays = None
        _keys = keys.copy()
        if self.extra_variables:
            # check if any of the extra variables are included in the expressions
            extra_vars = []
            for name in self.extra_variables:
                if name in keys:
                    extra_vars.append(name)
                    _keys.remove(name)
            if extra_vars:
                extra_arrays = {key: self.extra_variables[key] for key in extra_vars}

        tree_arrays = self.tree.arrays(_keys, library="ak", how=dict)
        if extra_arrays is not None:
            tree_arrays.update(extra_arrays)
        return tree_arrays

    @staticmethod
    def array_exporter(dict_of_arrays, **kwargs):
        library = kwargs.get("library", "ak")
        how = kwargs.get("how", dict)

        # TODO: long-term we want exporters + factory methods for these
        # e.g. {("ak", tuple): AKArrayTupleExporter, ("numpy", tuple): NumpyArrayTupleExporter}
        # reason: we want to be able to use these exporters with different libraries or
        # in different places in this codebase and allow to inject functionality for ranges and masks
        if library in LIBRARIES["awkward"]:
            if how == dict:
                return dict_of_arrays
            elif how == list:
                return [value for value in dict_of_arrays.values()]
            elif how == tuple:
                return tuple(value for value in dict_of_arrays.values())

        if library in LIBRARIES["pandas"]:
            return Uproot4Methods.arraydict_to_pandas(dict_of_arrays)

    def arrays(self, expressions, *args, **kwargs):
        if "outputtype" in kwargs:
            # renamed uproot3 -> uproot4
            outputtype = kwargs.pop("outputtype")
            kwargs["how"] = outputtype

        operations = kwargs.get("operations", [])
        tree_arrays = self.array_dict(keys=expressions)
        for operation in operations:
            for key, value in tree_arrays.items():
                tree_arrays[key] = operation(value)
        return self.array_exporter(tree_arrays, **kwargs)

    def array(self, key):
        return self[key]

    def evaluate(self, expression, **kwargs):
        return ak.numexpr.evaluate(expression, self, **kwargs)

    @staticmethod
    def counts(array, **kwargs):
        axis = kwargs.pop("axis", 1)
        return ak.count(Uproot4Methods.only_valid_entries(array), axis=axis, **kwargs)

    @staticmethod
    def all(array, **kwargs):
        axis = kwargs.pop("axis", 1)
        return ak.all(array, axis=axis, **kwargs)

    @staticmethod
    def pad(array, length: int, **kwargs: Dict[str, Any]) -> ak.Array:
        axis = kwargs.pop("axis", -1)
        return ak.pad_none(array, length, axis=axis, **kwargs)

    @staticmethod
    def flatten(array, **kwargs):
        return ak.flatten(array, **kwargs)

    @staticmethod
    def sum(array, **kwargs):
        axis = kwargs.pop("axis", None)
        return ak.sum(array, axis=axis, **kwargs)

    @staticmethod
    def prod(array, **kwargs):
        return ak.prod(array, **kwargs)

    @staticmethod
    def any(array, **kwargs):
        axis = kwargs.pop("axis", -1)
        return ak.any(array, axis=axis, **kwargs)

    @staticmethod
    def count_nonzero(array, **kwargs):
        return ak.count_nonzero(array, **kwargs)

    @staticmethod
    def max(array, **kwargs):
        axis = kwargs.pop("axis", 1)
        return ak.max(array, axis=axis, **kwargs)

    @staticmethod
    def min(array, **kwargs):
        return ak.min(array, **kwargs)

    @staticmethod
    def argmax(array, **kwargs):
        return ak.argmax(array, **kwargs)

    @staticmethod
    def argmin(array, **kwargs):
        return ak.argmin(array, **kwargs)

    @staticmethod
    def count_zero(array, **kwargs):
        return ak.count_zero(array, **kwargs)

    @staticmethod
    def dtype(array, **kwargs):
        t = ak.type(array).type
        if hasattr(t, "dtype"):
            return t.dtype
        return t

    @staticmethod
    def is_bool(array, **kwargs):
        t = Uproot4Methods.dtype(array)
        if "bool" in str(t):
            return True
        return Uproot4Methods.dtype(array) == "bool"

    @staticmethod
    def arrays_as_np_array(data, array_names, **kwargs):
        """
        Takes input data and converts it to an array of numpy arrays.
        e.g. arrays_as_np_array(data, ["x", "y"])
        results in
        array(<array of x>, <array of y>)
        """
        return data.arrays(
            array_names,
            library="ak",
            outputtype=list,
            **kwargs,
        )

    @staticmethod
    def to_pandas(data, keys):
        return data.arrays(
            keys,
            library="pd",
        )

    @staticmethod
    def valid_entry_mask(data: TreeLike) -> ak.Array:
        return ~ak.is_none(data)

    @staticmethod
    def only_valid_entries(data: TreeLike) -> ak.Array:
        return data[Uproot4Methods.valid_entry_mask(data)]

    @staticmethod
    def filtered_len(data: TreeLike) -> int:
        return len(data[~ak.is_none(data)])

    @staticmethod
    def fill_none(data: ArrayLike, fill_value, **kwargs):
        axis = kwargs.pop("axis", None)
        return ak.fill_none(array=data, value=fill_value, axis=axis, **kwargs)

    @staticmethod
    def values_as_type(data: ArrayLike, dtype, **kwargs):
        return ak.values_astype(data, dtype, **kwargs)


ArrayMethods = Uproot4Methods


class TreeToDictAdaptorV0(IndexingMixin, Uproot3Methods, TreeToDictAdaptor):
    pass


class TreeToDictAdaptorV1(IndexingMixin, Uproot4Methods, TreeToDictAdaptor):
    pass


register("uproot3", TreeToDictAdaptorV0)
register("uproot4", TreeToDictAdaptorV1)


# class ApplyRange(Callable):
#     def __init__(self, range):
#         self.range = range

#     def __call__(self, arrays):
#         pass

# class ApplyMask(Callable):
#     def __init__(self, mask):
#         self.mask = mask

#     def __call__(self, arrays):
#         pass

class Ranger(object):
    """
        TODO: range is just a different way of indexing --> refactor
    """
    tree: TreeToDictAdaptor
    start: int
    stop: int
    block_size: int
    mask: Any

    def __init__(self, tree: TreeToDictAdaptor, start: int, stop: int) -> None:
        self.tree = tree
        self.start = start
        self.stop = stop

        tree_size = tree.num_entries
        self.block_size = stop - start if stop > start > 0 else tree_size
        self.mask = np.ones(tree_size, dtype=bool)
        if self.block_size < tree_size:
            self.mask = np.zeros(tree_size, dtype=bool)
            self.mask[start:stop] = True

    @property
    def num_entries(self) -> int:
        """Returns the size of the range - overwrites tree.num_entries."""
        return self.block_size

    @property
    def unfiltered_num_entries(self) -> int:
        return self.tree.num_entries

    def __getitem__(self, key):
        return ak.mask(self.tree[key], self.mask)

    def __setitem__(self, key, value):
        # self.tree[key][self.start:self.stop] = value[self.start:self.stop]
        self.tree[key] = value

    def __delitem__(self, key):
        del self.tree[key]

    def __contains__(self, key):
        return key in self.tree

    def __len__(self):
        return self.block_size

    def array(self, key):
        return self[key]

    def arrays(self, *args, **kwargs):
        operations = kwargs.pop("operations", [])
        operations.append(lambda x: ak.mask(x, self.mask))
        kwargs["operations"] = operations
        arrays = self.tree.arrays(*args, **kwargs)
        return arrays

    def new_variable(self, name, value):
        import awkward as ak
        if len(value) < self.tree.num_entries:
            new_value = ak.concatenate(
                [
                    ak.Array([None] * self.start),
                    value,
                    ak.Array([None] * (self.tree.num_entries - self.stop))
                ],
                axis=0
            )
        else:
            new_value = value
        assert len(new_value) == self.tree.num_entries
        self.tree.new_variable(name, new_value, context=self.tree)

    def evaluate(self, expression, **kwargs):
        import awkward as ak
        return ak.numexpr.evaluate(expression, self, **kwargs)

    def keys(self):
        return self.tree.keys()

    def arrays_to_pandas(self, *args, **kwargs):
        return self.tree.arrays_to_pandas(*args, **kwargs)


def combine_masks(masks):
    import awkward as ak
    if len(masks) == 0:
        return ak.Array([])
    elif len(masks) == 1:
        return masks[0]
    else:
        return ak.concatenate(masks, axis=0)


class Masked(object):
    _mask: Any
    _tree: Ranger

    def __init__(self, tree: Ranger, mask: Any) -> None:
        self._tree = tree
        self._mask = np.ones(tree.num_entries, dtype=bool) if mask is None else mask

        if mask is not None and len(mask) < tree.unfiltered_num_entries:
            self._mask = ak.concatenate(
                [
                    ak.Array([False] * tree.start),
                    self._mask,
                    ak.Array([False] * (tree.unfiltered_num_entries - tree.stop))
                ],
                axis=0
            )

    def __getitem__(self, key):
        if self._mask is None:
            return self._tree[key]
        try:
            if len(self._mask) > len(self._tree):
                return self._tree[key][self._tree.start:self._tree.stop].mask[self._mask]
        except TypeError as e:
            raise e
        return self._tree[key].mask[self._mask]

    def __len__(self):
        return len(self._tree)

    def __contains__(self, key):
        return key in self._tree

    @property
    def num_entries(self) -> int:
        return self._tree.num_entries

    def count_nonzero(self):
        if self._mask is None:
            return len(self._tree)
        return ak.count_nonzero(self._mask)

    def apply_mask(self, mask):
        if self._mask is None:
            self._mask = mask
        else:
            self._mask = self._mask & mask

    def reset_mask(self):
        self._mask = None

    def reset_cache(self):
        pass

    def array(self, key):
        return self[key]

    def arrays(self, *args, **kwargs):
        operations = kwargs.pop("operations", [])
        if self._mask is not None:
            operations.append(lambda x: ak.mask(x, self._mask))

        kwargs["operations"] = operations
        arrays = self._tree.arrays(*args, **kwargs)
        return arrays

    def evaluate(self, expression, **kwargs):
        import awkward as ak
        return ak.numexpr.evaluate(expression, self, **kwargs)

    def keys(self):
        return self._tree.keys()

    def new_variable(self, name, value):
        self._tree.new_variable(name, value)


def create(arguments: Dict[str, Any]) -> TreeToDictAdaptor:
    """
    Create a TreeToDictAdaptor from a tree.
    """
    args_copy = arguments.copy()
    adapter_type = args_copy.pop("adapter", DEFAULT_TREE_TO_DICT_ADAPTOR)

    try:
        creation_func = adapters[adapter_type]
        return creation_func(**args_copy)
    except KeyError:
        raise ValueError(f"No adapter named {adapter_type}")


def create_ranged(arguments: Dict[str, Any]) -> Ranger:
    """
    Create a tree adapter with ranged access.
    """
    args_copy = arguments.copy()

    start = args_copy.pop("start", 0)
    stop = args_copy.pop("stop", -1)
    tree = create(args_copy)

    return Ranger(tree, start, stop)


def create_masked(arguments: Dict[str, Any]) -> Masked:
    """
    Create a tree adapter with masked access.
    """
    args_copy = arguments.copy()

    mask = args_copy.pop("mask", None)
    tree = create_ranged(args_copy)

    return Masked(tree, mask)


def create_masked_multitree(arguments: Dict[str, Any]) -> Masked:
    raise NotImplementedError("Multitree not yet implemented")



import pytest

import awkward as ak
import numpy as np


@pytest.mark.xfail(reason="sums are expected to differ")
def test_compare_np_to_ak(uproot4_tree):
    """ see https://github.com/scikit-hep/awkward-1.0/issues/1241 """
    with open('tests/data/eventweights.npy', 'rb') as f:
        np_array = np.load(f)

    np_sum = np.sum(np_array)
    for axis in [None, -1, 0]:
        ak_sum = ak.sum(np_array, axis=axis)
        assert ak_sum == np_sum



"""
Functions to run a job using alphatwirl
"""
from cachetools import cachedmethod
from cachetools.keys import hashkey
from functools import partial
import operator
from typing import Any, Dict

import awkward as awk
import numpy as np

from fast_carpenter.data_import import DataImportBase, get_data_import_plugin
from fast_carpenter.tree_adapter import create_masked


class BEvents(object):
    """
    from https://github.com/shane-breeze/atuproot/blob/master/atuproot/BEvents.py
    """
    non_branch_attrs = [
        "tree", "nevents_in_tree", "nevents_per_block", "nblocks",
        "start_block", "stop_block", "iblock", "start_entry", "stop_entry",
        "_branch_cache", "_nonbranch_cache", "size", "config",
    ]

    def __init__(
        self, tree, nevents_per_block=100000, start_block=0, stop_block=-1,
        branch_cache={},
    ):
        self.tree = tree
        self.nevents_in_tree = len(tree)
        self.nevents_per_block = int(nevents_per_block) \
            if nevents_per_block >= 0 \
            else self.nevents_in_tree

        nblocks = int((self.nevents_in_tree-1)/self.nevents_per_block + 1)
        start_block = min(nblocks, start_block)
        if stop_block > -1:
            self.nblocks = min(nblocks-start_block, stop_block)
        else:
            self.nblocks = nblocks-start_block
        self.stop_block = stop_block
        self.start_block = start_block
        self.iblock = -1

        self._branch_cache = branch_cache
        self._nonbranch_cache = {}

    def __len__(self):
        return self.nblocks

    def __repr__(self):
        return '{}({})'.format(
            self.__class__.__name__,
            self._repr_content(),
        )

    def _repr_content(self):
        return 'tree = {!r}, nevents_in_tree = {!r}, nevents_per_block = {!r}, '\
            'nblocks = {!r}, iblock = {!r}, start_block = {!r}, '\
               'stop_block = {!r}'.format(
                   self.tree,
                   self.nevents_in_tree,
                   self.nevents_per_block,
                   self.nblocks,
                   self.iblock,
                   self.start_block,
                   self.stop_block,
               )

    def __getitem__(self, i):
        if i >= self.nblocks:
            self.iblock = -1
            raise IndexError("The index is out of range: " + str(i))
        self._branch_cache.clear()

        self.iblock = i
        return self

    def __iter__(self):
        for self.iblock in range(self.nblocks):
            self._branch_cache.clear()
            yield self
        self.iblock = -1
        self._nonbranch_cache = {}

    def __getattr__(self, attr):
        if attr in self.non_branch_attrs:
            return getattr(self, attr)
        elif attr in self._nonbranch_cache:
            return self._nonbranch_cache[attr]
        return self._get_branch(attr)

    def __setattr__(self, attr, val):
        if attr in self.non_branch_attrs:
            super(BEvents, self).__setattr__(attr, val)
        else:
            if not (isinstance(val, awk.Array) or isinstance(val, np.ndarray)):
                self._nonbranch_cache[attr] = val
            else:
                key = hashkey('BEvents._get_branch', attr)
                self._branch_cache[key] = val

    @cachedmethod(operator.attrgetter('_branch_cache'), key=partial(hashkey, 'BEvents._get_branch'))
    def _get_branch(self, name):
        self.start_entry = (self.start_block + self.iblock) * self.nevents_per_block
        self.stop_entry = min(
            (self.start_block + self.iblock + 1) * self.nevents_per_block,
            (self.start_block + self.nblocks) * self.nevents_per_block,
            self.nevents_in_tree,
        )
        self.size = self.stop_entry - self.start_entry
        try:
            branch = "asdsd"
            print(branch)
            # branch = self.tree.array(
            #     name,
            #     entrystart=self.start_entry,
            #     entrystop=self.stop_entry,
            # )
        except KeyError as e:
            raise AttributeError(e)
        return branch

    def hasbranch(self, branch, encoding='utf-8'):
        return (
            branch.encode(encoding) in self.tree.keys()
            or hashkey('BEvents._get_branch', branch) in self._branch_cache
            or branch in self._nonbranch_cache
        )

    def delete_branches(self, branches):
        for branch in branches:
            key = hashkey('BEvents._get_branch', branch)
            if key in self._branch_cache:
                self._branch_cache.popitem(key)
            elif branch in self._nonbranch_cache:
                self._nonbranch_cache.popitem(branch)

    def __contains__(self, branch):
        return branch in getattr(self, "tree")


class EventRanger():
    def __init__(self):
        self._owner = None

    def set_owner(self, owner):
        self._owner = owner

    @property
    def start_entry(self):
        return (self._owner.start_block + self._owner.iblock) * self._owner.nevents_per_block

    @property
    def stop_entry(self):
        i_block = min(self._owner.iblock + 1, self._owner.nblocks)
        stop_entry = (self._owner.start_block + i_block) * self._owner.nevents_per_block
        return min(self._owner.nevents_in_tree, stop_entry)

    @property
    def entries_in_block(self):
        if self._owner and self._owner.iblock > -1:
            return self.stop_entry - self.start_entry
        return None


class BEventsWrapped(BEvents):
    def __init__(self, tree, *args, **kwargs):
        ranges = EventRanger()

        super(BEventsWrapped, self).__init__(tree, *args, **kwargs)
        ranges.set_owner(self)
        tree = create_masked(
            {
                "tree": tree,
                "start": self.start_entry,
                "stop": self.stop_entry,
                "adapter": "uproot4",
            }
        )
        self.tree = tree

    def _block_changed(self):
        self.tree.reset_mask()
        self.tree.reset_cache()

    def __getitem__(self, i):
        result = super(BEventsWrapped, self).__getitem__(self, i)
        self._block_changed()
        return result

    def __iter__(self):
        for value in super(BEventsWrapped, self).__iter__():
            self._block_changed()
            yield value
        self._block_changed()

    @property
    def start_entry(self):
        return (self.start_block + self.iblock) * self.nevents_per_block

    @property
    def stop_entry(self):
        i_block = min(self.iblock + 1, self.nblocks)
        stop_entry = (self.start_block + i_block) * self.nevents_per_block
        return min(self.nevents_in_tree, stop_entry)


class EventBuilder(object):
    data_import_plugin: DataImportBase = None

    def __init__(self, config):
        self.config = config

    def __repr__(self):
        return '{}({!r})'.format(
            self.__class__.__name__,
            self.config,
        )

    def __call__(self):
        if EventBuilder.data_import_plugin is None:
            EventBuilder.data_import_plugin = get_data_import_plugin("uproot4", None)
        data = EventBuilder.data_import_plugin.open(self.config.inputPaths)
        tree = data[self.config.treeName]

        events = BEventsWrapped(
            tree,
            self.config.nevents_per_block,
            self.config.start_block,
            self.config.stop_block,
        )
        events.config = self.config
        return events


class DummyCollector():
    def collect(self, *args, **kwargs):
        pass


class AtuprootContext:
    def __init__(self, plugins: Dict[str, Any] = None) -> None:
        self.plugins = plugins

    def __enter__(self):
        import atuproot.atuproot_main as atup
        self.atup = atup
        self._orig_event_builder = atup.EventBuilder
        self._orig_build_parallel = atup.build_parallel

        from atsge.build_parallel import build_parallel
        atup.EventBuilder = EventBuilder
        atup.EventBuilder.data_import_plugin = self.plugins["data_import"]
        atup.build_parallel = build_parallel
        return self

    def __exit__(self, *args, **kwargs):
        self.atup.EventBuilder = self._orig_event_builder
        self.atup.build_parallel = self._orig_build_parallel


def execute(sequence, datasets, args, plugins: Dict[str, Any] = None):
    """
    Run a job using alphatwirl and atuproot
    """

    if args.ncores < 1:
        args.ncores = 1

    sequence = [(s, s.collector() if hasattr(s, "collector") else DummyCollector()) for s in sequence]

    with AtuprootContext(plugins) as runner:
        process = runner.atup.AtUproot(
            args.outdir,
            quiet=args.quiet,
            parallel_mode=args.mode,
            process=args.ncores,
            max_blocks_per_dataset=args.nblocks_per_dataset,
            max_blocks_per_process=args.nblocks_per_sample,
            nevents_per_block=args.blocksize,
            profile=args.profile,
            profile_out_path="profile.txt",
        )

        ret_val = process.run(datasets, sequence)

    if not args.profile:
        # This breaks in AlphaTwirl when used with the profile option
        summary = {s[0].name: list(df.index.names) for s, df in zip(sequence, ret_val[0]) if df is not None}
    else:
        summary = " (Results summary not available with profile mode) "

    return summary, ret_val



import numpy as np
import six
from typing import List

from ..tree_adapter import ArrayMethods

__all__ = ["get_pandas_reduction"]


class BadReductionConfig(Exception):
    pass


class JaggedNth(object):
    def __init__(self, index, fill_missing, force_float=True):
        self.index = index
        self.fill_missing = fill_missing
        self.dtype = None
        if fill_missing is True or fill_missing is False:
            self.dtype = np.bool8
        elif force_float or isinstance(fill_missing, float):
            self.dtype = np.float64
        else:
            self.dtype = np.int32

    def __call__(self, array):
        padding = abs(self.index) + 1 if self.index >= 0 else abs(self.index)
        result = ArrayMethods.pad(array, padding)
        result = ArrayMethods.fill_none(result, self.fill_missing, axis=-1)
        if self.dtype is not None:
            result = ArrayMethods.values_as_type(result, self.dtype)
        return result[..., self.index]


class JaggedMethod(object):
    SUPPORTED: List[str] = ["sum", "prod", "any", "all", "count_nonzero",
                            "max", "min", "argmin", "argmax"]
    DEFAULTS = {
        "count_nonzero": {"axis": 1},
        "sum": {"axis": 1},
    }

    def __init__(self, method):
        self.method_name = method
        self._defaults = self.DEFAULTS.get(method, {})

    def __call__(self, array):
        return getattr(ArrayMethods, self.method_name)(array, **self._defaults)


class JaggedProperty(object):
    SUPPORTED: List[str] = ["counts"]

    def __init__(self, prop_name):
        self.prop_name = prop_name

    def __call__(self, array):
        return getattr(ArrayMethods, self.prop_name)(array)


def get_awkward_reduction(stage_name, reduction, fill_missing=np.nan):
    if isinstance(reduction, six.integer_types):
        return JaggedNth(int(reduction), fill_missing)

    if not isinstance(reduction, six.string_types):
        msg = "{}: requested reduce method is not a string or an int"
        raise BadReductionConfig(msg.format(stage_name))

    if reduction in JaggedMethod.SUPPORTED:
        return JaggedMethod(reduction)
    if reduction in JaggedProperty.SUPPORTED:
        return JaggedProperty(reduction)

    msg = "{}: Unknown method to reduce: '{}'"
    raise BadReductionConfig(msg.format(stage_name, reduction))


_numpy_ops = ["count_zero"]


class PandasAggregate(object):
    SUPPORTED: List[str] = ["sum", "prod", "max", "min", "argmax", "argmin"]

    def __init__(self, method):
        self.method = method

    def __call__(self, groups):
        return groups.agg(self.method)


class PandasNth(object):
    def __init__(self, index):
        self.index = index

    def __call__(self, groups):
        return groups.nth(self.index)


def get_pandas_reduction(stage_name, reduction):
    if not isinstance(reduction, (six.string_types, six.integer_types)):
        msg = "{}: requested reduce method is not a string or an int"
        raise BadReductionConfig(msg.format(stage_name))

    if reduction in PandasAggregate.SUPPORTED:
        return PandasAggregate(reduction)
    elif reduction in _numpy_ops:
        op = getattr(np, reduction)
        return PandasAggregate(op)
    else:
        try:
            index = int(reduction)
        except ValueError:
            pass
        else:
            return PandasNth(index)

    msg = "{}: Unknown method to reduce: '{}'"
    raise BadReductionConfig(msg.format(stage_name, reduction))



import pytest
from pytest_lazyfixture import lazy_fixture

from fast_carpenter import tree_adapter


@pytest.mark.parametrize(
    "tree, start, stop, expected_num_entries",
    [
        (lazy_fixture("input_tree"), 0, -1, lazy_fixture("size_of_test_sample")),
        (lazy_fixture("input_tree"), 0, lazy_fixture("size_of_test_sample"), lazy_fixture("size_of_test_sample")),
        (lazy_fixture("input_tree"), 100, 300, 200),
    ]
)
def test_ranged(tree, start, stop, expected_num_entries):
    test_tree = tree_adapter.create_ranged(
        {
            "adapter": "uproot4", "tree": tree,
            "start": start, "stop": stop,
        }
    )
    assert len(test_tree) == expected_num_entries



import awkward as ak
import numpy as np
import pytest
from pytest_lazyfixture import lazy_fixture

from fast_carpenter.testing import FakeBEEvent
import fast_carpenter.tree_adapter as tree_adapter
from fast_carpenter.tree_adapter import ArrayMethods


###############################################################################
# Uproot3 tests
###############################################################################


@pytest.fixture
def uproot3_adapter(uproot3_tree):
    return tree_adapter.create({"adapter": "uproot3", "tree": uproot3_tree})


def test_uproot3_num_entries(uproot3_tree, uproot3_adapter):
    assert uproot3_adapter.num_entries == uproot3_tree.numentries


def test_uproot3_getitem(uproot3_tree, uproot3_adapter):
    assert ak.all(uproot3_adapter["Muon_Py"] == uproot3_tree["Muon_Py"].array())

###############################################################################
# Uproot4 tests
###############################################################################


@pytest.fixture
def uproot4_adapter(uproot4_tree):
    return tree_adapter.create({"adapter": "uproot4", "tree": uproot4_tree})


@pytest.fixture
def uproot4_ranged_adapter(uproot4_tree, event_range):
    return tree_adapter.create_ranged(
        {
            "adapter": "uproot4",
            "tree": uproot4_tree,
            "start": event_range.start_entry,
            "stop": event_range.stop_entry
        }
    )


@pytest.fixture
def uproot4_masked_adapter(uproot4_tree, event_range):
    return tree_adapter.create_masked(
        {
            "adapter": "uproot4", "tree": uproot4_tree,
            "start": event_range.start_entry, "stop": event_range.stop_entry,
            "mask": [(i % 2) == 0 for i in range(event_range.start_entry, event_range.stop_entry)]
        }
    )


def test_uproot4_num_entries(uproot4_tree, uproot4_adapter):
    assert uproot4_adapter.num_entries == uproot4_tree.num_entries


def test_uproot4_getitem(uproot4_tree, uproot4_adapter):
    assert ak.all(uproot4_adapter["Muon_Py"] == uproot4_tree["Muon_Py"].array())


def test_uproot4_evaluate(uproot4_tree, uproot4_adapter):
    result = uproot4_adapter.evaluate("Muon_Py * NMuon")
    assert ak.num(result, axis=0) == ak.num(uproot4_tree["Muon_Py"].array(), axis=0)


def test_uproot4_range(uproot4_tree, uproot4_ranged_adapter, event_range):
    assert uproot4_ranged_adapter.num_entries == event_range.entries_in_block


def test_uproot4_add_retrieve(uproot4_tree, uproot4_ranged_adapter):
    muon_px = uproot4_ranged_adapter["Muon_Px"]
    assert ArrayMethods.filtered_len(muon_px) == len(uproot4_ranged_adapter)

    muon_py, muon_pz = uproot4_ranged_adapter.arrays(["Muon_Py", "Muon_Pz"], how=tuple)
    muon_momentum = np.hypot(muon_py, muon_pz)
    uproot4_ranged_adapter.new_variable("Muon_momentum", muon_momentum)

    retrieve_momentum = uproot4_ranged_adapter["Muon_momentum"]
    assert len(retrieve_momentum) == len(muon_momentum)
    assert ak.all(ak.flatten(retrieve_momentum) == ak.flatten(muon_momentum))


def test_overwrite(uproot4_ranged_adapter):
    muon_px = uproot4_ranged_adapter["Muon_Px"]
    assert ("Muon_Px" in uproot4_ranged_adapter)
    with pytest.raises(ValueError) as err:
        uproot4_ranged_adapter.new_variable("Muon_Px", muon_px / muon_px)
    assert "Muon_Px" in str(err)


def test_to_pandas(full_wrapped_tree):
    chunk = FakeBEEvent(full_wrapped_tree, "mc")
    inputs = ['Electron_Px', 'Electron_Py', 'EventWeight']
    df = ArrayMethods.to_pandas(chunk.tree, inputs)
    assert list(df.keys()) == inputs


def test_arraydict_to_pandas_with_new_variable(uproot4_ranged_adapter):
    muon_py, muon_pz = uproot4_ranged_adapter.arrays(["Muon_Py", "Muon_Pz"], how=tuple)
    muon_momentum = np.hypot(muon_py, muon_pz)
    uproot4_ranged_adapter.new_variable("Muon_momentum", muon_momentum)

    inputs = ['Muon_Py', 'Muon_Pz', 'Muon_momentum']
    arrays = {
        'Muon_Py': muon_py,
        'Muon_Pz': muon_pz,
        'Muon_momentum': muon_momentum,
    }
    df = ArrayMethods.arraydict_to_pandas(arrays)

    assert list(df.keys()) == inputs
    assert len(df) == ak.count_nonzero(muon_py)


@pytest.mark.parametrize(
    "tree_under_test",
    [
        lazy_fixture("uproot4_adapter"),
        lazy_fixture("uproot4_ranged_adapter"),
        lazy_fixture("uproot4_masked_adapter"),
    ]
)
def test_to_pandas_with_new_variable(tree_under_test):
    muon_py, muon_pz = tree_under_test.arrays(["Muon_Py", "Muon_Pz"], how=tuple)
    muon_momentum = np.hypot(muon_py, muon_pz)
    assert len(muon_momentum) == len(muon_py)
    tree_under_test.new_variable("Muon_momentum", muon_momentum)

    inputs = ['Muon_Py', 'Muon_Pz', 'Muon_momentum']
    df = ArrayMethods.to_pandas(tree_under_test, inputs)

    assert list(df.keys()) == inputs
    assert len(df) == ak.count_nonzero(muon_py)


@pytest.mark.parametrize(
    "tree_under_test, how",
    [
        (lazy_fixture("uproot4_adapter"), tuple),
        (lazy_fixture("uproot4_adapter"), list),
        (lazy_fixture("uproot4_ranged_adapter"), tuple),
        (lazy_fixture("uproot4_ranged_adapter"), list),
        (lazy_fixture("uproot4_masked_adapter"), tuple),
        (lazy_fixture("uproot4_masked_adapter"), list),
    ]
)
def test_arrays_to_tuple_or_list(tree_under_test, how):
    muon_py, muon_pz = tree_under_test.arrays(["Muon_Py", "Muon_Pz"], how=tuple)
    muon_momentum = np.hypot(muon_py, muon_pz)
    tree_under_test.new_variable("Muon_momentum", muon_momentum)
    _, _, muon_momentum_new = tree_under_test.arrays(["Muon_Py", "Muon_Pz", "Muon_momentum"], how=how)
    assert ak.all(muon_momentum_new == muon_momentum)


@pytest.mark.parametrize(
    "tree_under_test",
    [
        lazy_fixture("uproot4_adapter"),
        lazy_fixture("uproot4_ranged_adapter"),
        lazy_fixture("uproot4_masked_adapter"),
    ]
)
def test_arrays_to_dict(tree_under_test):
    muon_py, muon_pz = tree_under_test.arrays(["Muon_Py", "Muon_Pz"], how=tuple)
    muon_momentum = np.hypot(muon_py, muon_pz)
    tree_under_test.new_variable("Muon_momentum", muon_momentum)
    array_dict = tree_under_test.arrays(["Muon_Py", "Muon_Pz", "Muon_momentum"], how=dict)
    assert ak.all(array_dict["Muon_momentum"] == muon_momentum)


@pytest.mark.parametrize(
    "tree_under_test",
    [
        lazy_fixture("uproot4_adapter"),
        lazy_fixture("uproot4_ranged_adapter"),
        lazy_fixture("uproot4_masked_adapter"),
    ]
)
def test_arrays_as_np_lists(tree_under_test):
    muon_py, muon_pz = tree_under_test.arrays(["Muon_Py", "Muon_Pz"], how=tuple)
    muon_momentum = np.hypot(muon_py, muon_pz)
    tree_under_test.new_variable("Muon_momentum", muon_momentum)
    np_array = ArrayMethods.arrays_as_np_array(tree_under_test, ["Muon_Py", "Muon_Pz", "Muon_momentum"], how=dict)
    assert ak.all(np_array[-1] == muon_momentum)



"""
Functions to run a job using Coffea
"""
import copy
from fast_carpenter.tree_adapter import create_masked, TreeLike
from collections import namedtuple
from coffea import processor as cop
import logging


EventRanger = namedtuple("EventRanger", "start_entry stop_entry entries_in_block")
SingleChunk = namedtuple("SingleChunk", "tree config")
ChunkConfig = namedtuple("ChunkConfig", "dataset")
ConfigProxy = namedtuple("ConfigProxy", "name eventtype")


class CoffeaConnector(TreeLike):

    def __init__(self, data):
        self._data = data

    def __getattr__(self, name):
        if hasattr(self._data, name):
            return getattr(self._data, name)
        return getattr(self, name)

    def __getitem__(self, key):
        return self._data[key]

    @property
    def num_entries(self):
        return len(self._data)

    @property
    def dataset(self):
        return self._data.metadata["dataset"]

    @property
    def start(self):
        return self._data.metadata['entrystart']

    @property
    def stop(self):
        return self._data.metadata['entrystop']

    def arrays(self, keys, **kwargs):
        import awkward as ak
        library = kwargs.get("library", "ak")
        how = kwargs.get("how", dict)
        if library == "ak" and how == dict:
            return {key: ak.Array(self._data[key]) for key in keys}

        raise NotImplementedError(f"Cannot return arrays for {library=} and {how=}")

    def keys(self):
        return self._data.fields


class stages_accumulator(cop.AccumulatorABC):
    def __init__(self, stages):
        self._zero = copy.deepcopy(stages)
        self._value = copy.deepcopy(stages)

    def identity(self):
        return stages_accumulator(self._zero)

    def __getitem__(self, idx):
        return self._value[idx]

    def add(self, other):
        for i, stage in enumerate(self._value):
            if not hasattr(stage, "merge"):
                continue
            stage.merge(other[i])


class FASTProcessor(cop.ProcessorABC):
    def __init__(self, sequence):

        self._columns = list()
        self._sequence = sequence
        accumulator_dict = {'stages': cop.dict_accumulator({})}
        self._accumulator = cop.dict_accumulator(accumulator_dict)

    @property
    def columns(self):
        return self._columns

    @property
    def accumulator(self):
        return self._accumulator

    def process(self, df):
        output = self.accumulator.identity()
        connector = CoffeaConnector(df)

        # tree = MaskedUprootTree(df._tree, EventRanger(start, stop, stop - start))
        tree = create_masked(
            dict(
                tree=connector,
                start=connector.start,
                stop=connector.stop
            ))

        dsname = connector.dataset
        cfg_proxy = ConfigProxy(dsname, 'data' if dsname == 'data' else 'mc')
        chunk = SingleChunk(tree, ChunkConfig(cfg_proxy))

        output['stages'][dsname] = stages_accumulator(self._sequence)

        for work in output['stages'][dsname]._value:
            work.event(chunk)

        return output

    def postprocess(self, accumulator):
        stages = accumulator['stages']
        results = {}

        wf = copy.deepcopy(self._sequence)
        for i_step, step in enumerate(wf):
            if not hasattr(step, "collector"):
                continue
            collector = step.collector()
            output = collector.collect([(d, (s[i_step],)) for d, s in stages.items()])
            results[step.name] = output

        accumulator['results'] = results

        return accumulator


def load_execution_cfg(config):
    if not isinstance(config, str):
        return config

    import yaml
    with open(config, "r") as infile:
        cfg = yaml.safe_load(infile)
        return cfg


def configure_parsl(n_threads, monitoring, **kwargs):
    from parsl.config import Config
    from parsl.executors.threads import ThreadPoolExecutor
    from parsl.addresses import address_by_hostname

    if monitoring:
        from parsl.monitoring import MonitoringHub
        monitoring = MonitoringHub(hub_address=address_by_hostname(),
                                   hub_port=55055,
                                   logging_level=logging.INFO,
                                   resource_monitoring_interval=10,
                                   )
    else:
        monitoring = None

    local_threads = ThreadPoolExecutor(max_threads=n_threads,
                                       label='local_threads')
    config = Config(executors=[local_threads],
                    monitoring=monitoring,
                    strategy=None,
                    app_cache=True,
                    )
    return config


def configure_dask(client, **kwargs):
    from dask.distributed import Client

    if client:
        return Client(client)
    else:
        return Client(**kwargs)


def create_executor(args):
    exe_type = args.mode.split(":", 1)[-1].lower()
    exe_args = {}
    if getattr(args, "execution_cfg", None):
        exe_args = load_execution_cfg(args.execution_cfg)

    if exe_type == "local":
        executor = cop.futures_executor
        exe_args.setdefault('workers', args.ncores)
    elif exe_type == "parsl":
        executor = cop.parsl_executor
        exe_args.setdefault('n_threads', args.ncores)
        exe_args.setdefault('monitoring', False)
        if 'config' not in exe_args:
            exe_args['config'] = configure_parsl(**exe_args)
    elif exe_type == "dask":
        executor = cop.dask_executor
        exe_args.setdefault('processes', False)
        exe_args.setdefault('threads_per_worker', 2)
        exe_args.setdefault('n_workers', 2)
        exe_args.setdefault('memory_limit', '1GB')
        exe_args.setdefault('client', None)
        exe_args['client'] = configure_dask(**exe_args)
    else:
        msg = "Coffea executor not yet included in fast-carpenter: '%s'"
        raise NotImplementedError(msg % exe_type)

    return executor, exe_args


def execute(sequence, datasets, args, plugins):
    fp = FASTProcessor(sequence)

    executor, exe_args = create_executor(args)

    coffea_datasets = {}
    for ds in datasets:
        coffea_datasets[ds.name] = ds.__dict__.copy()
        coffea_datasets[ds.name].pop('name')
        coffea_datasets[ds.name]['treename'] = coffea_datasets[ds.name].pop('tree')

    maxchunks = args.nblocks_per_dataset
    if maxchunks < 1:
        maxchunks = None
    out = cop.run_uproot_job(coffea_datasets, 'events', fp,
                             executor, executor_args=exe_args,
                             chunksize=args.blocksize, maxchunks=maxchunks)

    return out["stages"], out["results"]
